import imageio

from mouselungseg import LungsPredictor, extract_3d_roi

if __name__ == "__main__":
    image = imageio.imread("https://zenodo.org/record/8099852/files/lungs_ct.tif")
    print(image.shape)

    predictor = LungsPredictor()

    mask = predictor.predict(image)
    print(mask.shape, mask.sum())

    roi, mask_roi = extract_3d_roi(image, mask)
    print(roi.shape, mask_roi.shape, mask_roi.sum())



import os
import sys

import napari.layers
from napari.qt.threading import thread_worker
from PyQt5.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QGridLayout,
    QLabel,
    QProgressBar,
    QPushButton,
    QWidget,
    QSizePolicy,
)

sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
)

from mouselungseg.predict import LungsPredictor

class LungsSegmentationWidget(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()

        self.viewer = napari_viewer
        self.predictor = LungsPredictor()

        # Layout
        grid_layout = QGridLayout()
        grid_layout.setAlignment(Qt.AlignTop)
        self.setLayout(grid_layout)

        # Image
        self.cb_image = QComboBox()
        self.cb_image.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        grid_layout.addWidget(QLabel("Image", self), 0, 0)
        grid_layout.addWidget(self.cb_image, 0, 1)

        # Compute button
        self.btn = QPushButton("Segment lungs", self)
        self.btn.clicked.connect(self._start_segmentation)
        grid_layout.addWidget(self.btn, 3, 0, 1, 2)

        # Progress bar
        self.pbar = QProgressBar(self, minimum=0, maximum=1)
        self.pbar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        grid_layout.addWidget(self.pbar, 4, 0, 1, 2)

        # Setup layer callbacks
        self.viewer.layers.events.inserted.connect(
            lambda e: e.value.events.name.connect(self._on_layer_change)
        )
        self.viewer.layers.events.inserted.connect(self._on_layer_change)
        self.viewer.layers.events.removed.connect(self._on_layer_change)
        self._on_layer_change(None)

    def _on_layer_change(self, e):
        self.cb_image.clear()
        for x in self.viewer.layers:
            if isinstance(x, napari.layers.Image):
                if x.data.ndim in [2, 3]:
                    self.cb_image.addItem(x.name, x.data)

    @thread_worker
    def _tumor_prediction_thread(self):
        return self.predictor.predict(self.selected_image)

    def _start_segmentation(self):
        self.selected_image = self.cb_image.currentData()
        if self.selected_image is None:
            return

        self.pbar.setMaximum(0)

        worker = self._tumor_prediction_thread()
        worker.returned.connect(self._load_in_viewer)
        worker.start()

    def _load_in_viewer(self, segmentation):
        """Callback from thread returning."""
        if segmentation is not None:
            segmentation_layer = self.viewer.add_labels(segmentation, name="Lungs (Yolo)")
            segmentation_layer.opacity = 0.2
            segmentation_layer.blending = "additive"
        
        self.pbar.setMaximum(1)



import pandas as pd
from skimage.measure import regionprops_table


def extract_3d_roi(image, labels):
    df = pd.DataFrame(
        regionprops_table(
            labels, 
            intensity_image=image, 
            properties=['bbox', 'image']
        )
    )

    x0 = int(df["bbox-0"].values[0])
    x1 = int(df["bbox-3"].values[0])
    y0 = int(df["bbox-1"].values[0])
    y1 = int(df["bbox-4"].values[0])
    z0 = int(df["bbox-2"].values[0])
    z1 = int(df["bbox-5"].values[0])

    roi = image[x0:x1, y0:y1, z0:z1]
    roi_mask = df['image'][0]

    return roi, roi_mask


import os
import numpy as np
from skimage.transform import resize
from skimage.exposure import rescale_intensity
import pooch
import scipy.ndimage as ndi
from ultralytics import YOLO

from mouselungseg.postprocess import extract_3d_roi


MODEL_PATH = os.path.expanduser(
    os.path.join(os.getenv("XDG_DATA_HOME", "~"), ".mousetumornet")
)


def retreive_model():
    """Downloads the model weights from Zenodo."""
    pooch.retrieve(
        url="https://zenodo.org/records/13268683/files/yolo_seg_mouselungs.pt",
        known_hash="md5:139471da545565d033748dc0d54a2392",
        path=MODEL_PATH,
        progressbar=True,
        fname="yolo_seg_mouselungs.pt",
    )


def to_rgb(arr):
    return np.repeat(arr[..., None], repeats=3, axis=-1)


def handle_2d_predict(image, model, imgsz):
    image = rescale_intensity(image, out_range=(0, 255)).astype(np.uint8)
    image = to_rgb(image)

    results = model.predict(
        source=image,
        conf=0.25,  # Confidence threshold for detections.
        iou=0.5,  # Intersection over union threshold.
        imgsz=imgsz,  # Square resizing
        max_det=2,  # Two detections max
        augment=False,
    )

    mask = np.zeros_like(image, dtype=np.uint16)
    r = results[0]
    if r.masks is not None:
        mask = r.masks.cpu().numpy().data[0]  # First mask only
        mask = resize(mask, image.shape, order=0) == 1
        mask[mask] = 1

        # Keep one of the channels only
        if len(mask.shape) == 3:
            mask = mask[..., 0]

        # Fill-in the mask
        mask = ndi.binary_fill_holes(
            mask, structure=ndi.generate_binary_structure(2, 1)
        )

    if len(mask.shape) == 3:
        mask = mask[..., 0]

    return mask


def handle_3d_predict(image, model, imgsz):
    n_slices = len(image)

    mask_3d = []
    for slice_idx, z_slice in enumerate(image):
        print(f"{slice_idx} / {n_slices}")
        mask_2d = handle_2d_predict(z_slice, model, imgsz)
        mask_3d.append(mask_2d)

    mask_3d = np.stack(mask_3d)

    # Dilate in the Z direcion to suppress missing frames
    mask_3d = ndi.binary_dilation(
        mask_3d, structure=ndi.generate_binary_structure(3, 1), iterations=2
    )

    return mask_3d


def handle_predict(image, model, imgsz):
    if len(image.shape) == 2:
        mask = handle_2d_predict(image, model, imgsz)
    elif len(image.shape) == 3:
        mask = handle_3d_predict(image, model, imgsz)

    mask = mask.astype(np.uint8)

    return mask


class LungsPredictor:
    def __init__(self):
        retreive_model()

        self.model = YOLO(os.path.join(MODEL_PATH, "yolo_seg_mouselungs.pt"))
        self.imgsz = 640

    def predict(self, image: np.ndarray) -> np.ndarray:
        mask = handle_predict(image, self.model, self.imgsz)
        return mask
    
    def compute_3d_roi(self, image: np.ndarray) -> np.ndarray:
        mask = self.predict(image)
        roi, roi_mask = extract_3d_roi(image, mask)
        return roi, roi_mask



from mouselungseg import LungsPredictor
import tifffile
from pathlib import Path
import argparse
import glob


def process_input_file_predict(input_image_file, predictor):
    image = tifffile.imread(input_image_file)

    segmentation = predictor.predict(image)

    pt = Path(input_image_file)
    out_file_name = pt.parent / f"{pt.stem}_mask.tif"

    tifffile.imwrite(out_file_name, segmentation)

    print("Wrote to ", out_file_name)


def cli_predict_image():
    """Command-line entry point for model inference."""
    parser = argparse.ArgumentParser(description="Use this command to run inference.")
    parser.add_argument(
        "-i",
        type=str,
        required=True,
        help="Input image. Must be a TIF image file.",
    )
    args = parser.parse_args()

    input_image_file = args.i

    predictor = LungsPredictor()

    process_input_file_predict(input_image_file, predictor)


def cli_predict_folder():
    parser = argparse.ArgumentParser(
        description="Use this command to run inference in batch on a given folder."
    )
    parser.add_argument(
        "-i",
        type=str,
        required=True,
        help="Input folder. Must contain suitable TIF image files.",
    )
    args = parser.parse_args()

    input_folder = args.i

    lungs_predict = LungsPredictor()

    for input_image_file in glob.glob(str(Path(input_folder) / "*.tif")):
        process_input_file_predict(input_image_file, lungs_predict)
